"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
function convertFlowDeclarationToRunnableDeclaration(preparedFlowieExecution, isAsyncFunction, isGeneratorFunction) {
    const initialSubFlows = [];
    const initialMainFlow = {
        functionsFromContainers: [],
        steps: []
    };
    const { mainFlow, subFlows } = preparedFlowieExecution.flows
        .reduce(readFlowsAndSubFlows, {
        subFlows: initialSubFlows,
        mainFlow: initialMainFlow,
        isAsyncFunction,
        isGeneratorFunction,
        generatorsCount: 0
    });
    const runnableDeclaration = {
        isAsync: preparedFlowieExecution.isAsync,
        mainFlow,
        subFlows
    };
    return runnableDeclaration;
}
exports.default = convertFlowDeclarationToRunnableDeclaration;
function readFlowsAndSubFlows(readFlowsAndSubFlowsReducer, flowElement, currentIndex, array) {
    const { isAsyncFunction, isGeneratorFunction, mainFlow, subFlows, generatorsCount } = readFlowsAndSubFlowsReducer;
    const { mainFlow: newMainFlow, subFlows: newSubFlows, generatorsFound } = new FlowElementReader(isAsyncFunction, isGeneratorFunction, mainFlow, subFlows).read(flowElement);
    const totalOfGenerators = generatorsCount + generatorsFound;
    if (currentIndex === array.length - 1 && totalOfGenerators > 0) {
        return {
            generatorsCount: totalOfGenerators,
            mainFlow: {
                ...newMainFlow,
                steps: newMainFlow.steps.concat({ finishGeneratorsCount: totalOfGenerators })
            },
            subFlows: newSubFlows,
            isAsyncFunction: readFlowsAndSubFlowsReducer.isAsyncFunction,
            isGeneratorFunction: readFlowsAndSubFlowsReducer.isGeneratorFunction
        };
    }
    return {
        generatorsCount: totalOfGenerators,
        mainFlow: newMainFlow,
        subFlows: newSubFlows,
        isAsyncFunction: readFlowsAndSubFlowsReducer.isAsyncFunction,
        isGeneratorFunction: readFlowsAndSubFlowsReducer.isGeneratorFunction
    };
}
class FlowElementReader {
    constructor(isAsyncFunction, isGeneratorFunction, mainFlow, subFlows) {
        this.isAsyncFunction = isAsyncFunction;
        this.isGeneratorFunction = isGeneratorFunction;
        this.mainFlow = mainFlow;
        this.subFlows = subFlows;
        this.readSplitItem = this.readSplitItem.bind(this);
    }
    read(flowElement) {
        const pipeFlow = flowElement;
        if (pipeFlow.pipe)
            return this.readPipeStep(pipeFlow);
        const splitFlow = flowElement;
        if (splitFlow.split)
            return this.readSplitStep(splitFlow);
        const preparedFlowie = flowElement;
        return this.readSubFlow(preparedFlowie);
    }
    readPipeStep(pipeFlow) {
        if (typeof pipeFlow.pipe === 'string') {
            const functionName = pipeFlow.pipe;
            const isGenerator = this.isGeneratorFunction(functionName);
            const step = this.createStepForFunction(functionName, isGenerator);
            return {
                mainFlow: {
                    functionsFromContainers: getUniqueFunctionNames(this.mainFlow.functionsFromContainers, functionName),
                    steps: this.mainFlow.steps.concat(step)
                },
                subFlows: this.subFlows,
                generatorsFound: Number(isGenerator)
            };
        }
        return this.readSubFlow(pipeFlow.pipe);
    }
    readSplitStep(splitFlow) {
        const initialSplitStepReading = {
            functionsFromContainers: [],
            subFlows: [],
            splitStep: {
                isAsync: false,
                split: []
            }
        };
        const { functionsFromContainers, subFlows, splitStep } = splitFlow.split.reduce(this.readSplitItem, initialSplitStepReading);
        return {
            generatorsFound: 0,
            mainFlow: {
                functionsFromContainers: getUniqueFunctionNames(this.mainFlow.functionsFromContainers, ...functionsFromContainers),
                steps: this.mainFlow.steps.concat(splitStep)
            },
            subFlows: this.subFlows.concat(subFlows)
        };
    }
    readSplitItem(splitStepReading, flowieItemDeclaration) {
        const { functionsFromContainers, subFlows, splitStep } = splitStepReading;
        if (typeof flowieItemDeclaration === 'string') {
            const isGenerator = this.isGeneratorFunction(flowieItemDeclaration);
            const isAsync = this.isAsyncFunction(flowieItemDeclaration);
            if (isGenerator) {
                const hash = `generator_on_split_${flowieItemDeclaration}`;
                return {
                    functionsFromContainers,
                    subFlows: subFlows.concat({
                        isAsync,
                        functionsFromContainers: [flowieItemDeclaration],
                        hash,
                        steps: [{ generator: flowieItemDeclaration, isAsync }, { finishGeneratorsCount: 1 }]
                    }),
                    splitStep: {
                        isAsync: keepPreviousTrue(splitStep.isAsync, isAsync),
                        split: splitStep.split.concat({ flow: hash, isAsync })
                    }
                };
            }
            return {
                functionsFromContainers: getUniqueFunctionNames(functionsFromContainers, flowieItemDeclaration),
                subFlows,
                splitStep: {
                    isAsync: keepPreviousTrue(splitStep.isAsync, this.isAsyncFunction(flowieItemDeclaration)),
                    split: splitStep.split.concat(flowieItemDeclaration)
                }
            };
        }
        const { flowStep, subFlows: newSubFlows } = new SubFlowElementReader(this.isAsyncFunction, this.isGeneratorFunction)
            .read(flowieItemDeclaration);
        return {
            functionsFromContainers,
            subFlows: subFlows.concat(newSubFlows),
            splitStep: {
                isAsync: keepPreviousTrue(splitStep.isAsync, flowStep.isAsync),
                split: splitStep.split.concat(flowStep)
            }
        };
    }
    readSubFlow(flowElement) {
        const { flowStep, subFlows } = new SubFlowElementReader(this.isAsyncFunction, this.isGeneratorFunction)
            .read(flowElement);
        return {
            generatorsFound: 0,
            mainFlow: {
                functionsFromContainers: this.mainFlow.functionsFromContainers,
                steps: this.mainFlow.steps.concat(flowStep)
            },
            subFlows: this.subFlows.concat(subFlows)
        };
    }
    createStepForFunction(functionName, isGenerator) {
        const isAsync = this.isAsyncFunction(functionName);
        if (isGenerator) {
            return { generator: functionName, isAsync };
        }
        return { pipe: functionName, isAsync };
    }
}
class SubFlowElementReader {
    constructor(isAsyncFunction, isGeneratorFunction) {
        this.isAsyncFunction = isAsyncFunction;
        this.isGeneratorFunction = isGeneratorFunction;
        this.readSubFlowSteps = this.readSubFlowSteps.bind(this);
        this.readSubFlowSplitItem = this.readSubFlowSplitItem.bind(this);
    }
    read(flowElement) {
        const { name } = flowElement;
        const hash = name || generateHash();
        const { isAsync, functionsFromContainers, steps, subFlows } = this.readSubFlow(flowElement);
        const subFlow = { isAsync, functionsFromContainers, hash, steps };
        return {
            flowStep: { flow: hash, isAsync },
            subFlows: [subFlow, ...subFlows]
        };
    }
    readSubFlow(flowElement) {
        const pipeFlow = flowElement;
        if (pipeFlow.pipe) {
            return this.readSubFlowPipe(pipeFlow);
        }
        const splitFlow = flowElement;
        if (splitFlow.split) {
            return this.readSubFlowSplit(splitFlow);
        }
        const preparedFlowie = flowElement;
        const initialFlowReading = {
            isAsync: false,
            functionsFromContainers: [],
            steps: [],
            subFlows: [],
            generatorsCount: 0
        };
        const subFlowReading = preparedFlowie.flows.reduce(this.readSubFlowSteps, initialFlowReading);
        const { generatorsCount } = subFlowReading;
        return {
            ...subFlowReading,
            steps: subFlowReading.steps.concat(generatorsCount ? { finishGeneratorsCount: generatorsCount } : [])
        };
    }
    readSubFlowPipe(pipeFlow) {
        if (typeof pipeFlow.pipe === 'string') {
            const functionName = pipeFlow.pipe;
            const isAsync = this.isAsyncFunction(functionName);
            const isGenerator = this.isGeneratorFunction(functionName);
            const step = this.createStepForFunctionForSubFlow(functionName, isGenerator);
            return {
                isAsync,
                functionsFromContainers: [functionName],
                steps: [step],
                subFlows: [],
                generatorsCount: Number(isGenerator)
            };
        }
        const { flowStep, subFlows } = new SubFlowElementReader(this.isAsyncFunction, this.isGeneratorFunction)
            .read(pipeFlow.pipe);
        return {
            isAsync: flowStep.isAsync,
            functionsFromContainers: [],
            steps: [flowStep],
            subFlows,
            generatorsCount: 0
        };
    }
    readSubFlowSplit(splitFlow) {
        const initialFlowReading = {
            isAsync: false,
            functionsFromContainers: [],
            steps: [{ split: [], isAsync: false }],
            subFlows: [],
            generatorsCount: 0
        };
        return splitFlow.split.reduce(this.readSubFlowSplitItem, initialFlowReading);
    }
    readSubFlowSplitItem(subflowReading, flowieItemDeclaration) {
        const { isAsync, functionsFromContainers, steps, subFlows } = subflowReading;
        const [splitStep] = steps;
        if (typeof flowieItemDeclaration === 'string') {
            const isGenerator = this.isGeneratorFunction(flowieItemDeclaration);
            const willBeAsync = keepPreviousTrue(isAsync, this.isAsyncFunction(flowieItemDeclaration));
            if (isGenerator) {
                const hash = `generator_on_split_${flowieItemDeclaration}`;
                return {
                    isAsync: willBeAsync,
                    functionsFromContainers,
                    steps: [{
                            split: splitStep.split.concat({ flow: hash, isAsync: willBeAsync }),
                            isAsync: willBeAsync
                        }],
                    subFlows: subFlows.concat({
                        isAsync,
                        functionsFromContainers: [flowieItemDeclaration],
                        hash,
                        steps: [{ generator: flowieItemDeclaration, isAsync }, { finishGeneratorsCount: 1 }]
                    }),
                    generatorsCount: 0
                };
            }
            return {
                isAsync: willBeAsync,
                functionsFromContainers: getUniqueFunctionNames(functionsFromContainers, flowieItemDeclaration),
                steps: [{
                        split: splitStep.split.concat(flowieItemDeclaration),
                        isAsync: willBeAsync
                    }],
                subFlows,
                generatorsCount: 0
            };
        }
        const { flowStep, subFlows: collectedSubFlows } = new SubFlowElementReader(this.isAsyncFunction, this.isGeneratorFunction).read(flowieItemDeclaration);
        const willBeAsync = keepPreviousTrue(isAsync, flowStep.isAsync);
        return {
            isAsync: willBeAsync,
            functionsFromContainers,
            steps: [{
                    split: splitStep.split.concat(flowStep),
                    isAsync: willBeAsync
                }],
            subFlows: subFlows.concat(collectedSubFlows),
            generatorsCount: 0
        };
    }
    readSubFlowSteps(subFlowReading, flowElement) {
        const { isAsync, steps, subFlows, functionsFromContainers, generatorsCount } = this.readSubFlowItem(flowElement);
        return {
            isAsync: keepPreviousTrue(subFlowReading.isAsync, isAsync),
            functionsFromContainers: getUniqueFunctionNames(subFlowReading.functionsFromContainers, ...functionsFromContainers),
            steps: subFlowReading.steps.concat(steps),
            subFlows: subFlowReading.subFlows.concat(subFlows),
            generatorsCount: generatorsCount + subFlowReading.generatorsCount
        };
    }
    readSubFlowItem(flowElement) {
        const preparedFlowie = flowElement;
        if (preparedFlowie.flows) {
            const { flowStep, subFlows } = new SubFlowElementReader(this.isAsyncFunction, this.isGeneratorFunction)
                .read(flowElement);
            return {
                isAsync: flowStep.isAsync,
                functionsFromContainers: [],
                steps: [flowStep],
                subFlows,
                generatorsCount: 0
            };
        }
        return new SubFlowElementReader(this.isAsyncFunction, this.isGeneratorFunction).readSubFlow(flowElement);
    }
    createStepForFunctionForSubFlow(functionName, isGenerator) {
        const isAsync = this.isAsyncFunction(functionName);
        if (isGenerator) {
            return { generator: functionName, isAsync };
        }
        return { pipe: functionName, isAsync };
    }
}
function getUniqueFunctionNames(functionNames, ...newValues) {
    return Array.from(new Set(functionNames.concat(newValues)));
}
function keepPreviousTrue(previous, next) {
    return previous || next;
}
function generateHash() {
    return Math.random().toString(36).slice(2);
}
//# sourceMappingURL=convertFlowDeclarationToRunnableDeclaration.js.map